#include "test/servo_runner.h"

#include <emlabcpp/protocol/streams.h>

using namespace schpin;

void servo_runner::visitor::to_servo( s_ctl_id msg_id, reply r, const master_lewan_message& m )
{
        SCHPIN_ASSERT( runner.ser_ );

        SCHPIN_INFO_LOG( "servo", "To servo: " << m );
        runner.ser_->write( m );
        std::this_thread::sleep_for( std::chrono::milliseconds( 50 ) );
        if ( r == reply::NO ) {
                return;
        }
        runner.ser_->read_with_sequencer< lewan_sequencer< lewan_master_message > >().match(
            [&]( lewan_master_message m ) {
                    SCHPIN_INFO_LOG( "servo", "From servo: " << m );
                    lewan_master_extract( m ).match(
                        [&]( std::tuple<
                             lewan_header_start,
                             lewan_header_start,
                             lewan_servo_id,
                             lewan_master_variant,
                             lewan_checksum > pack ) {
                                auto     clk_time = get_system_time();
                                servo_id sid      = *std::get< lewan_servo_id >( pack );
                                runner.servos_[sid]->on_servo_receive(
                                    clk_time, msg_id, std::get< lewan_master_variant >( pack ) );
                        },
                        [&]( auto err ) {
                                SCHPIN_ERROR_LOG( "lewan_servo", "Lewan prot error: " << err );
                        } );
            },
            [&]( auto err ) {
                    SCHPIN_ERROR_LOG( "control", "Read error: " << err );
            } );
}

void servo_runner::visitor::to_servo( s_ctl_id msg_id, reply, const master_dynmx_message& m )
{
        SCHPIN_ASSERT( runner.ser_ );

        SCHPIN_INFO_LOG( "servo", "To servo: " << m );
        runner.ser_->write( m );

        runner.ser_->read_with_sequencer< dynmx_sequencer< dynmx_master_message > >()
            .convert_right( []( auto e ) {
                    return to_string( e );
            } )
            .bind_left( [&]( dynmx_master_message msg ) {
                    SCHPIN_INFO_LOG( "servo", "From servo: " << msg );
                    return dynmx_master_extract( msg ).convert_right( []( auto e ) {
                            return to_string( e );
                    } );
            } )
            .match(
                [&]( std::tuple< dynmx_header, dynmx_servo_id, dynmx_master_subvalue, dynmx_crc >
                         pack ) {
                        auto                        tpl = std::get< dynmx_master_subvalue >( pack );
                        const dynmx_response_error& err = std::get< 1 >( tpl );
                        if ( err != 0 ) {
                                SCHPIN_ERROR_LOG(
                                    "runner",
                                    "Dynmx return error: " << std::bitset< 8 >{ err }
                                                           << " from message: " << m );
                        }
                        const dynmx_buffer& buffer = std::get< 2 >( tpl );

                        servo_id sid = *std::get< dynmx_servo_id >( pack );
                        if ( buffer.size() != 0 ) {
                                runner.servos_[sid]->on_servo_receive(
                                    get_system_time(), msg_id, servo_receive_variant{ buffer } );
                        }
                },
                [&]( auto err ) {
                        SCHPIN_ERROR_LOG( "servo_runner", err );
                } );
}

void servo_runner::tick()
{
        visitor vis{ *this };
        for ( auto& [id, s] : servos_ ) {
                s->tick( get_system_time(), vis );
                s->poll( get_system_time(), 0, vis );
        }
        std::string msg = repl_.prompt();
        vm_.load( msg ).optionally_log();
        vm_.full_exec( [&]( lua_State* lua_state, auto event ) {
                   return handle_event( lua_state, event );
           } )
            .optionally_log();
}
void servo_runner::run_file( const std::filesystem::path& file )
{
        auto opt_e = vm_.load( file );
        if ( opt_e ) {
                opt_e.optionally_log();
                return;
        }
        bool finished = false;

        visitor vis{ *this };
        while ( !finished ) {
                for ( auto& [id, s] : servos_ ) {
                        s->tick( get_system_time(), vis );
                        s->poll( get_system_time(), 0, vis );
                }
                vm_.exec( [&]( lua_State* lua_ptr, auto event ) {
                        return handle_event( lua_ptr, event );
                } )
                    << match(
                           [&]( lua_res res ) {
                                   finished = res == lua_res::FINISHED;
                           },
                           [&]( slua_error err ) {
                                   finished = true;
                                   SCHPIN_ERROR_LOG( "runner", "Slua error: " << err );
                           } );
        }
}
void servo_runner::handle_event( lua_State*, print_help_event )
{
        nlohmann::json res = typename slua_runner_def::def{};
        print_help( std::cout, res );
}
void servo_runner::handle_event( lua_State*, print_config_event e )
{
        servos_[e.sid]->dump_info();
}
nlohmann::json servo_runner::handle_event( lua_State* lua_ptr, msg_event e )
{
        nlohmann::json res;

        // TODO: write proper thing for this in schpin_lib/json and check in rest of the
        // codebase
        try {
                auto var = e.j.get< master_servo_variant >();
                SCHPIN_ASSERT( servos_.contains( e.sid ) );
                servos_[e.sid]
                    ->on_servo_packet( ctl_id{ 0 }, var )
                    .match(
                        [&]( servo_events_variant var ) {
                                visitor vis{ *this };
                                servos_[e.sid]->on_event(
                                    ctl_id{ 0 }, get_system_time(), var, vis );
                        },
                        [&]( servo_master_variant var ) {
                                res = var;
                        } );
        }
        catch ( const nlohmann::json::exception& ex ) {
                lua_error_stream( lua_ptr, "Json error: ", ex.what() );
        }
        return res;
}
